# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

import bouillon


def test_version():
    assert(bouillon.__version__)


def test_name():
    assert(bouillon.__name__ == "bouillon")


def test_run():
    bouillon.run(["ls"] )
    bouillon.run(["ls"], dry_run=True)
    bouillon.run(["unknown"], dry_run=True)


def test_check_for_test_files_fail(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    a = src / "a.py"
    a.write_text("a")

    foo = src / "foo"
    foo.mkdir()
    b = foo / "b.py"
    b.write_text("b")

    test = tmp_path / "test"
    test.mkdir()
    test_a = test / "test_a.py"
    test_a.write_text("test_a")

    assert not bouillon.check_for_test_files(src, test)


def test_check_for_test_files(tmp_path):
    src = tmp_path / "src"
    src.mkdir()
    test = tmp_path / "test"
    test.mkdir()
    assert bouillon.check_for_test_files(src, test)

    a = src / "a.py"
    foo = src / "foo"
    foo.mkdir()
    b = foo / "b.py"
    a.write_text("a")
    b.write_text("b")

    test_a = test / "test_a.py"
    test_foo = test / "foo"
    test_foo.mkdir()
    test_b = test_foo / "test_b.py"
    test_a.write_text("test_a")
    test_b.write_text("test_b")

    assert bouillon.check_for_test_files(src, test)


def test_find_requirements():
    pass



# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

"""Various helpers.

Contains various helper functions that are intended to be usefull when writing
scripts for managing (build, test, release etc.) a project.
"""

from __future__ import annotations

import logging
import shutil
import subprocess
from pathlib import Path

logger = logging.getLogger(__name__)


def run(
    args: list[str],
    *,
    dry_run: bool = False,
    shell: bool = False,
    **kwargs,
) -> subprocess.CompletedProcess:
    """Run a command.

    Wrapper around subprocess.run, kwargs are forwarded to subprocess.run,
    dry_run = True, do not execute commands that make changes.
    """
    if shell:
        logger.warning("setting shell to True can cause problems.")

    logger.info(f' executing (dry-run={dry_run}): {" ".join(args)}')

    if shutil.which(args[0]) is None:
        logger.error(f'Command "{args[0]}" was not found.')

    if dry_run:
        return subprocess.CompletedProcess(
            args, 2, "dry-run output", "dry-run error")

    return subprocess.run(args, shell=shell, **kwargs)


def check_for_test_files(
    src_path: str,
    test_path: str,
    *,
    prefix: str = "test_",
    suffix: str = ".py",
    ignore: list[str] = ["__init__.py"],
) -> bool:
    """Check for test files.

    Check that all source files, all files in src_path with the defined suffix,
    have a correponding test file in the test_path with the defined prefix and
    suffix.
    """
    assert Path(src_path).exists(), f"path does not exist {src_path}"
    assert Path(test_path).exists(), f"path does not exist {test_path}"

    # Find all source files
    srcs = Path(src_path).rglob(f"*{suffix}")
    relative_srcs = list(map(lambda s: s.relative_to(src_path), srcs))
    if len(relative_srcs) == 0:
        logger.warning("No source files found.")

    # Find all test files
    tests = Path(test_path).rglob(f"{prefix}*{suffix}")
    relative_tests = map(lambda t: t.relative_to(test_path), tests)

    # Remove all tests files from the list of source files
    for t in relative_tests:
        relative_srcs.remove(Path(str(t).replace(prefix, "")))

    if len(relative_srcs) == 0:
        return True

    logger.warning(f"Missing tests for files: {relative_srcs}")
    return False



# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

import subprocess
import sys

import pytest

from bouillon.cli import cli, release

"""
We do a dry run test of some of our commands to verify that the cli basically
works but we do not want to excute the commands, that should be done in the
test pipeline. This helps us to avoid situation where we mess up the arguments
in our cli or some other trivial mistakes.
"""


def test_cli():
    assert cli(sys.argv[1:])

    with pytest.raises(SystemExit):
        assert cli(["release", "--help"])

    a = vars(cli(["release", "1.2.3"]))
    assert a["check_branch"]
    assert a["releaseable_branch"] in ["main", None]  # None on github runners
    assert a["distribution_dir"] == "dist"
    assert a["news_files"] == ["NEWS.rst",]
    assert a["build_steps"] == [["python", "-m", "build"],]
    assert a["lint_steps"] == [["brundle"], ["licensecheck", "--zero"]]
    assert a["test_steps"] == [["pytest"],]


def test_cli_news_files():
    a = vars(cli(["release", "100", "--news-files", "foo", "bar", "--news-files", "foobars"]))
    assert a["news_files"] == ["foo", "bar", "foobars"]

    a = vars(cli(["release", "100", "--news-files"]))
    assert a["news_files"] == []


def test_cli_lint_steps():
    a = vars(cli(["release", "100", "--lint-steps", "foo", "bar", "--lint-steps", "foo"]))
    assert a["lint_steps"] == [["foo", "bar"],["foo"]]


def test_cli_test_steps():
    a = vars(cli(["release", "100", "--test-steps", "foobar"]))
    assert a["test_steps"] == [["foobar"],]


def test_cli_build_steps():
    a = vars(cli(["release", "100", "--build-steps", "barfoo"]))
    assert a["build_steps"] == [["barfoo"],]


def test_release_invalid_version():
    with pytest.raises(SystemExit):
        release(**vars(cli(["--dry-run", "release", "fo0"])))


def test_release_existing_version():
    with pytest.raises(SystemExit):
        release(**vars(cli(["--dry-run", "release", "1.0.0"])))


def test_release_from_disallowed_branch():
    with pytest.raises(SystemExit):
        release(**vars(cli(["--dry-run", "release", "100.0.0",
                            "--releaseable-branch", "foo",])))


@pytest.mark.cicd
def test_release_unclean_branch():
    """Note fails if branch is behind remote or not clean."""
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*"])))


def test_release_unclean_branch_ok():
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*", "--check-branch"])))


def test_release_append_news_file():
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*", "--check-branch", "--news-files", "pyproject.toml"])))


def test_release_append_build_step():
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*", "--check-branch", "--build-steps", "build"])))


def test_release_append_lint_step():
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*", "--check-branch", "--lint-steps", "brundle"])))


def test_release_append_test_step():
    release(**vars(cli(["--dry-run", "release", "100.0.0",
        "--releaseable-branch", "*", "--check-branch", "--test-steps", "pytest"])))


def test_boil_help():
    subprocess.run(["bouillon"], check=True)


def test_boil_infile():
    subprocess.run(["bouillon", "-i", "pyproject.toml"], check=True)


def test_boil_build():
    subprocess.run(["bouillon", "--dry-run", "build"], check=True)


def test_boil_release_existing_version():
    with pytest.raises(Exception):
        subprocess.run(["bouillon", "--dry-run", "release", "1.0.0"], check=True)


def test_boil_release_invalid_version():
    with pytest.raises(Exception):
        subprocess.run(["bouillon", "--dry-run", "release", "foo3"], check=True)


def test_boil_clean():
    subprocess.run(["bouillon", "clean"], check=True)



# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

"""Git checks, non of them changes state."""

from __future__ import annotations

from pathlib import Path
from subprocess import PIPE, STDOUT, CalledProcessError

from bouillon import run


def repository_name() -> str:
    """Get git repository name."""
    r = run(["git", "config", "--get", "remote.origin.url"],
            stdout=PIPE,
            check=True)

    return Path(r.stdout.decode().rstrip()).stem


def current_branch() -> str:
    """Get git current branch."""
    r = run(["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stdout=PIPE,
            check=True)

    return r.stdout.decode().rstrip()


def default_branch() -> str | None:
    """Get git default branch, returns None if it cannot be determined."""
    try:
        r = run(["git", "rev-parse", "--abbrev-ref", "origin/HEAD"],
            stdout=PIPE,
            check=True)
        return r.stdout.decode().rstrip().lstrip("origin/")

    except CalledProcessError:
        return None


def working_directory_clean() -> bool:
    """Check if the working directory is clean."""
    try:
        run(["git", "diff", "--quiet", "--exit-code"],
            stdout=PIPE,
            check=True)
        return True

    except CalledProcessError:
        return False


def working_directory_updated() -> bool:
    """Check if the local working directory is up to date with the remote."""
    try:
        r = run(["git", "fetch", "--dry-run"],
            stdout=PIPE,
            stderr=STDOUT,
            check=True)
    except CalledProcessError:
        return False

    return r.stdout.decode().rstrip() == ""


def commit_id() -> str:
    """Get current git commit id."""
    r = run(["git", "rev-parse", "HEAD"],
            stdout=PIPE,
            check=True)

    return r.stdout.decode().rstrip()


def tags() -> list[str]:
    """Get list of all git tags."""
    r = run(["git", "tag"],
            stdout=PIPE,
            check=True)

    return r.stdout.decode().rstrip()



#! /usr/bin/env python3
#
# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

"""Command Line Interface (CLI) for project interaction.

Run various commands, such as; test, build, release on your project. You should
modify the steps that are relevant for your project, and the cli such that it
reflects those steps. The cli specified here is used for the bouillon module.
"""

from __future__ import annotations

import os
import shutil
import sys
from argparse import (
    ArgumentDefaultsHelpFormatter, ArgumentParser, FileType, Namespace,
)
from logging import basicConfig, getLevelName, getLogger
from typing import Callable, List

from packaging.version import InvalidVersion, Version

try:
    from tomllib import load  # type: ignore
except ModuleNotFoundError:
    from tomli import load  # type: ignore

from bouillon import git, run

logger = getLogger(__name__)


def build(*, build_steps: List[List[str]], dry_run: bool, **kwargs) -> None:
    """Build distributeables."""
    logger.info("Building source and binary distributions")
    [run(step, dry_run=dry_run) for step in build_steps]


def clean(*, distribution_dir: str, dry_run: bool, **kwargs) -> None:
    """Remove files and dirs created during build."""
    logger.info(f"Deleting {distribution_dir} directories.")
    if not dry_run:
        shutil.rmtree(distribution_dir, ignore_errors=True)


def release(
    *,
    check_branch: bool,
    releaseable_branch: str,
    version: str,
    distribution_dir: str,
    news_files: List[str],
    lint_steps: List[List[str]],
    test_steps: List[List[str]],
    dry_run: bool,
    **kwargs
) -> None:
    """Release the project."""
    try:
        if str(Version(version)) in git.tags():
            logger.error("Tag already exists.")
            exit(1)
    except InvalidVersion:
        logger.error("Provided version is not a valid version identifier")
        exit(1)

    logger.info("Running lint and test steps")
    clean(distribution_dir=distribution_dir, dry_run=dry_run, **kwargs)
    [run(step, dry_run=dry_run, check=True) for step in lint_steps]
    [run(step, dry_run=dry_run, check=True) for step in test_steps]

    if releaseable_branch not in ["*", git.current_branch()]:
        logger.error(f"Only release from the default branch {releaseable_branch}")
        exit(1)

    if check_branch:
        if not git.working_directory_clean():
            logger.error("Unstaged changes in the working directory.")
            exit(1)
        if not git.working_directory_updated():
            logger.error("Branch is behind remote.")
            exit(1)

    logger.info("Opening the news file(s) for edit and make a commit.")
    EDITOR = os.environ.get("EDITOR", "nano")
    [run([EDITOR, file], dry_run=dry_run) for file in news_files]
    [run(["git", "add", file], dry_run=dry_run) for file in news_files]
    run(["git", "commit", "-m", f"preparing release {version}"], dry_run=dry_run)

    logger.info("Create an annotated tag, used by setuptools_scm.")
    run(["git", "tag", "-a", f"{version}", "-m",
        f"creating tag {version} for new release"], dry_run=dry_run)

    build(dry_run=dry_run, **kwargs)

    logger.info("upload builds to pypi and push commit and tag to repo.")
    try:
        run(["twine", "upload", f"{distribution_dir}/*"], dry_run=dry_run, check=True)
    except Exception as e:
        logger.error(f"Upload failed with error {e}, cleaning")
        run(["git", "tag", "-d", f"{version}"], dry_run=dry_run)
        run(["git", "reset", "--hard", "HEAD~1"], dry_run=dry_run)
        exit(1)

    run(["git", "push"], dry_run=dry_run)
    run(["git", "push", "origin", f"{version}"], dry_run=dry_run)


class input_overwrites(list):
    """Avoid copying default argument(s), when appending inputs."""
    def __copy__(self):
        return []


def cli(args) -> Namespace:
    """Build the cli."""
    parser = ArgumentParser(
        description="Bouillon",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-i",
        "--infile",
        default="pyproject.toml",
        type=FileType("rb"),
        help="path to input file",
    )

    bouillon_settings = load(
        parser.parse_args(["-i", "pyproject.toml"]).infile).get(
            "tool", dict()).get("bouillon", dict())
#
    subparsers = parser.add_subparsers(help="available sub commands")

    parser_build = subparsers.add_parser("build", help="build.",
        formatter_class=ArgumentDefaultsHelpFormatter,)
    parser_build.set_defaults(function=build)
    parser_build.add_argument(
        "--build-steps", type=str, help="build steps.",
        nargs="*", action="append",
        default=input_overwrites(bouillon_settings.get("build_steps", [["python", "-m", "build"],])))

    parser_clean = subparsers.add_parser("clean", help="clean temp files.",
        formatter_class=ArgumentDefaultsHelpFormatter,)
    parser_clean.set_defaults(function=clean)
    parser_clean.add_argument(
        "--distribution-dir", type=str, help="distribution directory.",
        default=bouillon_settings.get("distribution_dir", "dist"))

    parser_release = subparsers.add_parser("release", help="release me.",
        formatter_class=ArgumentDefaultsHelpFormatter,
        description="""
        1. Check that the choosen tag is valid and does not already exists.
        2. Cleans the distribution folder.
        3. Run all linters.
        4. Run tests.
        5. Check that we are ok to release from current branch.
        6. Check that there are no unstaged changes on the current branch.
        7. Check that the current branch is not behind the remote.
        8. Opens all news files for editing.
        9. Add and commit all news files.
        10. Creates the tag.
        11. Build the project.
        12. Uploads to pypi.
        13. Push the commit and tag to the origin.

        Note that precedence of settings in decreasing order is as follows:
        commandline arguments -> project file (pyproject.toml) -> defaults.
        """
        )
    parser_release.add_argument("version", type=str,
                                help="release version (e.g. '1.2.3').")
    parser_release.add_argument(
        "--check-branch", action="store_false",
        help="check that the branch is clean and up to date with remote.",
        default=bouillon_settings.get("check_branch", True))
    parser_release.add_argument(
        "--releaseable-branch", type=str,
        help="branches from which release is allowed ('*' for any branch)",
        default=bouillon_settings.get("releaseable_branch", git.default_branch()))
    parser_release.add_argument(
        "--distribution-dir", type=str, help="distribution directory.",
        default=bouillon_settings.get("distribution_dir", "dist"))
    parser_release.add_argument(
        "--news-files", type=str, help="news files to open for edits.",
        nargs="*", action="extend",
        default=input_overwrites(bouillon_settings.get("news_files", ["NEWS.rst",])))
    parser_release.add_argument(
        "--build-steps", type=str, help="build steps.",
        nargs="*", action="append",
        default=input_overwrites(bouillon_settings.get("build_steps", [["python", "-m", "build"],])))
    parser_release.add_argument(
        "--lint-steps", type=str, help="lint steps.",
        nargs="*", action="append",
        default=input_overwrites(bouillon_settings.get("lint_steps", [["brundle"], ["licensecheck", "--zero"]])))
    parser_release.add_argument(
        "--test-steps", type=str, help="list of test steps.",
        nargs="*", action="append",
        default=input_overwrites(bouillon_settings.get("test_steps", [["pytest"],])))

    parser_release.set_defaults(function=release)

    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICIAL"],
        default="WARNING", help="set log level.")
    parser.add_argument(
        "--log-file", type=str, help="set log file.")
    parser.add_argument(
        "--dry-run", action="store_true",
        help="perform a dry run, its helpfull to also set the log-level.")

    def _print_help(**kwargs):
        parser.print_help()

    parser.set_defaults(check=True, function=_print_help)
    return parser.parse_args(args)


def main(*, function: Callable, log_level: str, log_file: str, **kwargs) -> None:
    """Setup logging and run a step."""
    basicConfig(
        filename=log_file,
        level=getLevelName(log_level),
    )
    logger.info(f"Running {function.__name__} step, with the argumnts: {kwargs}")
    function(**kwargs)


def main_cli() -> None:
    main(**vars(cli(sys.argv[1:])))


if __name__ == "__main__":
    main_cli()



# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

import pytest

from bouillon import git


def test_repository_name():
    assert git.repository_name() == "bouillon"


@pytest.mark.cicd
def test_working_directory_clean():
    """Note will fail if there are unstaged changes in the working dir."""
    assert git.working_directory_clean()


@pytest.mark.cicd
def test_working_directory_updated():
    """Note will fail if not up to date with origin."""
    assert git.working_directory_updated()


def test_current_branch():
    assert isinstance(git.current_branch(), str)


def test_default_branch():
    """28/09/24 Note that this return None on a gitub runner."""
    assert git.default_branch() in ["main", None]


def test_commit_id():
    assert len(git.commit_id()) == 40


def test_tags():
    assert "0.0.1" in git.tags()
    assert "1.0.0" in git.tags()
    assert "1.4.0" in git.tags()



# Copyright (c) 2020, Janus Heide.
# All rights reserved.
#
# Distributed under the "BSD 3-Clause License", see LICENSE.txt.

"""Import bouillon functions."""


from importlib.metadata import PackageNotFoundError, version

from bouillon.bouillon import check_for_test_files, run
from bouillon.git import (
    commit_id, current_branch, default_branch, repository_name, tags,
    working_directory_clean,
)

try:
    __version__ = version("bouillon")
except PackageNotFoundError:
    pass

__all__ = ("run",
           "check_for_test_files",
           "repository_name",
           "working_directory_clean",
           "current_branch",
           "default_branch",
           "commit_id",
           "tags")
